#!/usr/bin/perl -w
# 
# Generate graphs of the average PPD per WU in the queue
# 

use strict;
use Cwd;
use File::Basename;
use Date::Calc;
use Date::Manip;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";
use GD::Graph::bars;
use GD::Graph::lines;

# set the Date::Calc language
Date::Calc::Language(Date::Calc::Decode_Language('English'));

my $www_data_dir     = '/var/www/fci/';
my $client_data_dir  = 'client-data/';
my $xml_data_dir     = 'xml-data/';
my $queue_graphs_dir = 'queue-graphs/';

my %cfg = ( self                     => 'fci-generate-queue-graphs.pl',
            version                  => '0.1',
	    
	    verbose                  => 0,
	    quiet                    => 0,
	    
	    'www-data-dir'           => $www_data_dir,
	    'client-data-dir'        => $client_data_dir,
	    'xml-data-dir'           => $xml_data_dir,
            
            'client-list-file'       => $xml_data_dir.'client-list.xml',
	    'queue-graphs-file'      => $xml_data_dir.'queue-graphs.xml',

	    'action'                 => '',
	    'client'                 => '',
          );


# Get commandline parameters
Getopt::Mixed::init("dir=s
                     client=s
                     verbose
		     quiet
		     help
		     d>dir
		     c>client
		     v>verbose
		     q>quiet
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit 0;
	}
	
	$cfg{'www-data-dir'} = $value if($option eq 'dir');
	$cfg{'client'}       = $value if($option eq 'client');
	
	$cfg{verbose}    = 1      if($option eq 'verbose');
	$cfg{quiet}      = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# if quiet was requested together with verbose, disable verbose which contradicts quiet
if($cfg{quiet} && $cfg{verbose}) {
	#print "--quiet overrides --verbose\n";
	$cfg{verbose} = 0;
}

# add directory separator to the end of the www-data-dir path if missing
$cfg{'www-data-dir'} .= '/' if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');

my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}


# show what's print()ed without waiting for the newline $|=1;
$|=1;


# check file existance and permissions
# TODO

# read the xml files
my $client_list         = read_xml(%cfg, Name => 'client-list')    || die "Error: Cannot read client-list XML file!\n";
my $queue_graphs        = read_xml(%cfg, Name => 'queue-graphs')   || die "Error: Cannot read queue-graphs XML file!\n";

#print Dumper($config);
#print Dumper($queue_graphs);

if($cfg{client} eq '') {
	# For each clients processed qd-data.xml file, 
	# generate a graph with the average PPD for the WUs in the queue.dat
	
	print "All clients mode.\n\n" if($cfg{verbose});

	foreach my $client (keys %{$client_list->{client_list}->{client}}) {
	        next if($client !~ /^[a-zA-Z0-9\_\-]+$/);
		
		print "Processing client: $client\n" if($cfg{verbose});
	
		if(has_qd_data($client, $client_list, %cfg)) {
			generate_ppd_graph_for_client($client, %cfg);

			generate_flops_graph_for_client($client, %cfg);
		}
		else {
			print "No usable qd-data.xml for client: $client, skipping this client.\n" if($cfg{verbose});
			next;
		}
	
		print "\n" if($cfg{verbose});
	}
}
else {
	print "Single client mode.\n\n" if($cfg{verbose});
	
	if($cfg{client} =~ /^[a-zA-Z0-9\_\-]+$/) {
		print "Processing client: $cfg{client}\n" if($cfg{verbose});
	
		if(has_qd_data($cfg{client}, $client_list, %cfg)) {
			generate_ppd_graph_for_client($cfg{client}, %cfg);
			
			generate_flops_graph_for_client($cfg{client}, %cfg);
			
			if(has_fahlog($cfg{client}, $client_list, %cfg)) {
				generate_tpf_graph_for_client($cfg{client}, %cfg);
			}
		}
		else {
			print "Error: No usable qd-data.xml for client: $cfg{client}\n";
			exit 1;
		}
	}
}

print "Saving updated queue-graphs XML file ...\n" if($cfg{verbose});
my ($success, $reason) = write_xml( %cfg, 
                                    Name => 'queue-graphs',
				    XML  => $queue_graphs
				  );

if(!$success) {
	print "Error: Failed to write queue-graphs XML file!\n";
	exit 1;
}


exit 0; # exit normally

################################[ Subroutines ]################################

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}


##############
#
#     Routine: read_file()
#  Parameters: path to file
#     Returns: content of the file
# 	       undef in case of error
# Description: return the contents of a file
#

sub read_file {
	my $file    = $_[0];
	my $verbose = $_[1] ? $_[1] : 0;
	
	print "Reading file from disk: $file ... " if($verbose);
	if(-r $file) {
		my $data = '';

		open(F, "< $file") || return undef;
		while(<F>) {
			$data .= $_;
		}
		close F;

		print "Done.\n" if($verbose);

		return $data;
	}
	else {
		print "FAILED!\n" if($verbose);
		return undef;
	}
}


##############
#
#     Routine: save_file()
#  Parameters: path to file
# 	       data to save into the file
# 	       optional hash with
# 	         mode    => file premission in octal
# 	         binary  => flag to set binary mode (only needed on Windows)
# 	         verbose => flag to set verbose output
#     Returns: 1 for successfull save or undef for unsuccessfull save
#              the reason for the failure or an empty string
# Description: write a file to disk if the required permissions are set
#

sub save_file {
	my $filename = shift;
	my $content  = shift;
	my %option   = @_;

	my ($success, $reason) = (1, '');

	print "Saving file to disk: $filename ... " if($option{verbose});
	if(!-e $filename || -w $filename) {
		print "Done.\n" if($option{verbose});
		
		open(F, "> $filename") || return (0, "Error: Cannot save file: $filename ($!)\n"); 
		if($option{binary} && $option{binary} == 1) {
			print "Binmode enabled.\n" if($option{verbose});
			binmode(F);
		}
		print F $content;
		close F;

		if($option{mode} && $option{mode} =~ /^0\d{3}$/) {
			print "Set permission: $option{mode}.\n" if($option{verbose});
			chmod oct($option{mode}), $filename;
		}
		else {
			if($option{verbose}) {
				print "Ignoring invalid mode: ";
				print $option{mode};
				print "\n";
			}
		}

		return 1;
	}
	else {
		print "FAILED! (File is not writeable)\n" if($option{verbose});
		return (0, "Error: File does not exists or is unwriteable!\n");
	}
}


##############
#
#     Routine: resize_image()
#  Parameters: source image
#              target image
#              width in pixels
#              height in pixels
#     Returns: 1 on success
#              undef in case of error
# Description: Resize an image with Image::Magick
#

sub resize_image {
        my $source_file  = $_[0];
        my $target_file  = $_[1];
	my $width  	 = $_[2];
	my $height 	 = $_[3];
	my $blur   	 = 1;
	
	if(-r $source_file && (!-e $target_file || -w $target_file)) {
		my $image = Image::Magick->new(magick => 'jpg');

		$image->Read($source_file);

		my ($img_width, $img_height) = $image->Get('width','height');

		if($img_width > $img_height) {
			# Landscape orientated image

			if($img_width > $width) {
				# image is bigger that target size, so resize it
				
				my $new_width  = $width;
				my $new_height = $img_height / ($img_width / $new_width);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		elsif($img_height > $img_width) {
			# portrait orientated image
			
			if($img_height > $height) {
				# image is bigger that target size, so resize it

				my $new_height = $height;
				my $new_width  = $img_width / ($img_height / $new_height);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		else {
			# square image (width eq heigth)
			
			if($img_height > $height) {
				# image is bigger that target size, so resize it

				my $new_height = $height;
				my $new_width  = $img_width / ($img_height / $new_height);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
			elsif($img_width > $width) {
				# image is bigger that target size, so resize it
				
				 my $new_width  = $width;
				 my $new_height = $img_height / ($img_width / $new_width);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
			else {
				# image is bigger that target size, so resize it

				my $new_width  = $width;
				my $new_height = $img_height / ($img_width / $new_width);
				
				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		
		# write resized file to disk
		$image->Write( filename => $target_file );
		chmod(0664, $target_file);
		
		return 1;
	}
	else {
		return undef;
	}
}

##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
                elsif($args{Name} eq 'jmol-missing-projects') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'project' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
                elsif($args{Name} eq 'queue-graphs') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'client' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd)$/) {
		if(!$args{File}) {
			return (0, "No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-missing-projects') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
                elsif($args{Name} eq 'queue-graphs') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Unknown file: $args{Name}\n";
		}
	}

	return ($success, $reason);
}

sub get_ppd {
	my $current_index = shift;
	my $qd_data       = shift;
	
	my %info;
	if(exists $qd_data->{queue}->{index}->{"$current_index"}->{project_number} &&
	   $qd_data->{queue}->{index}->{"$current_index"}->{project_number} =~ /^\d+$/
	) {
		$info{project_number} = $qd_data->{queue}->{index}->{"$current_index"}->{project_number};

		#print "Found project number: $info{project_number}\n" if($cfg{verbose});
		
		# get run, clone & generation numbers
		if(exists $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{run} &&
		   exists $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{clone} &&
		   exists $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{generation} &&
		   $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{run} =~ /^\d+$/ &&
		   $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{clone} =~ /^\d+$/ &&
		   $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{generation} =~ /^\d+$/
		) {
			$info{run} = $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{run};
			$info{clone} = $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{clone};
			$info{generation} = $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{generation};
		
			#print "Found Run, Clone & Generation number: $info{run}, $info{clone}, $info{generation}\n" if($cfg{verbose});
		
			# get (credit|points)_per_hour and convert to points_per_day
			if(exists $qd_data->{queue}->{index}->{"$current_index"}->{credit_per_hour} && 
			   $qd_data->{queue}->{index}->{"$current_index"}->{credit_per_hour} =~ /^\d+\.\d+$/
			) {
				$info{ppd} = $qd_data->{queue}->{index}->{"$current_index"}->{credit_per_hour} * 24;
				
				#print "Found PPD: $info{ppd}\n" if($cfg{verbose});
			}
			elsif(exists $qd_data->{queue}->{index}->{"$current_index"}->{points_per_hour} &&
			      $qd_data->{queue}->{index}->{"$current_index"}->{points_per_hour} =~ /^\d+\.\d+$/
			) {
				$info{ppd} = $qd_data->{queue}->{index}->{"$current_index"}->{points_per_hour} * 24;
				
				#print "Found PPD: $info{ppd}\n" if($cfg{verbose});
			}
		}
	}

	return %info;
}

sub generate_ppd_graph_for_client {
	my $client_name = shift;
	my %cfg         = @_;

	my $qd_file = $client_list->{client_list}->{client}->{$client_name}->{files}->{qd_data};

        # qd-data.xml content
	print "Parsing qd-data XML file: $qd_file ... \n" if($cfg{verbose});
        my $qd_data = read_xml( %cfg, 
				Name => 'qd',
				File => $qd_file
			      ) || die "Failed to read project-images XML file!\n";

	#print Dumper($qd_data);
	
	my $current_index = -1;
	if($qd_data->{queue}->{current_index} && $qd_data->{queue}->{current_index} ne '') {
		$current_index = $qd_data->{queue}->{current_index};
	}
	
	# Start with the oldest WU
	if($current_index == 9) {
		$current_index = 0;
	}
	else {
		$current_index += 1;
	}
	
	my %project_ppd;
	my @project = ();
	my @ppd     = ();
	my $i       = 0;
	for(my $index = $current_index ; $index <= 9; $index++) {
		if(exists $qd_data->{queue}->{index}->{"$index"}->{project_number} &&
		   $qd_data->{queue}->{index}->{"$index"}->{project_number} =~ /^\d+$/
		) {
			%project_ppd = get_ppd($index, $qd_data);
			
			my $prcg  = sprintf("%4d", $project_ppd{project_number});
			   $prcg .= " ($project_ppd{run}, $project_ppd{clone}, $project_ppd{generation})";
		
			if($project_ppd{ppd}) {
				if($cfg{verbose}) {
					print "Adding PPD: ";
					printf("%8s", $project_ppd{ppd});
					print "  for: $prcg\n";
				}
			   
				push @project, $prcg;
				push @ppd, sprintf("%.2f", $project_ppd{ppd});
			}
		}
	}
	for(my $index = 0; $index < $current_index; $index++) {
		if(exists $qd_data->{queue}->{index}->{"$index"}->{project_number} &&
		   $qd_data->{queue}->{index}->{"$index"}->{project_number} =~ /^\d+$/
		) {
			%project_ppd = get_ppd($index, $qd_data);
		
			my $prcg  = sprintf("%4d", $project_ppd{project_number});
			   $prcg .= " ($project_ppd{run}, $project_ppd{clone}, $project_ppd{generation})";
			
			if($project_ppd{ppd}) {
				if($cfg{verbose}) {
					print "Adding PPD: ";
					printf("%8s", $project_ppd{ppd});
					print "  for: $prcg\n";
				}
			   
				push @project, $prcg;
				push @ppd, sprintf("%.2f", $project_ppd{ppd});
			}
		}
	}

	my $graph_file = $queue_graphs_dir.$client_name.'.png';
	my $width      = 500;   # GoT Max: 570
	my $height     = 400;   # GoT Max: 430

	my @x = @project;
	my @y = @ppd;
	my @l = @project; # legend

	my $max = 0;
	foreach(@ppd) {
		if($_ > $max) {
			$max = $_;
		}
	}
	$max = int($max + 50);

	if(@x && @y) {
		my $date = get_date();
	
	        my @data  = (\@x, \@y);
	        my $graph = GD::Graph::bars->new($width, $height);
	
	        my $data = GD::Graph::Data->new(\@data);
       		my $values = $data->copy;
		   
		   $graph->set(show_values => $values);
		   $graph->set(dclrs => [ qw(black) ] );
		   
		   $graph->set( x_label           => "Project (Run, Clone, Generation)",
		   		y_label           => "PPD",
				title             => "Average PPD Of Work Units In The Queue",
				x_labels_vertical => 1,
				y_max_value       => $max,
				
				fgclr             => '#000000',	
				bgclr             => '#FFFFFF',	
				labelclr          => '#000000',
				axislabelclr      => '#000000',
				legendclr         => '#000000',
				valuesclr         => '#000000',
				textclr           => '#000000',
			      ) or die $graph->error;
			
		my $gd = $graph->plot(\@data) or die $graph->error;
		
		my ($success, $reason) = save_file( $graph_file, 
		                                    $graph->plot(\@data)->png, 
						    mode => "0664", verbose => $cfg{verbose}
						  );
		if($success) {
			# add the graph file for this client to the queue-graphs list
			$queue_graphs->{queue_graphs}->{client}->{$client_name}->{file} = $graph_file;
			$queue_graphs->{queue_graphs}->{client}->{$client_name}->{last_update} = $date;

			print "Successfully generated graph for client: $client_name\n" if($cfg{verbose});
		}
		else {
			print $reason;
			exit 1;
		}
	}
	else {
		print "No data found to generate graph for client: $client_name.\n" if($cfg{verbose});
	}
}

sub generate_flops_graph_for_client {
	my $client_name = shift;
	my %cfg         = @_;

	my $qd_file = $client_list->{client_list}->{client}->{$client_name}->{files}->{qd_data};

        # qd-data.xml content
	print "Parsing qd-data XML file: $qd_file ... \n" if($cfg{verbose});
        my $qd_data = read_xml( %cfg, 
				Name => 'qd',
				File => $qd_file
			      ) || die "Failed to read project-images XML file!\n";

	#print Dumper($qd_data);
	
	my $current_index = -1;
	if($qd_data->{queue}->{current_index} && $qd_data->{queue}->{current_index} ne '') {
		$current_index = $qd_data->{queue}->{current_index};
	}
	
	# Start with the oldest WU
	if($current_index == 9) {
		$current_index = 0;
	}
	else {
		$current_index += 1;
	}
	
	my %project_ppd;
	my @project = ();
	my @ppd     = ();
	my $i       = 0;
	for(my $index = $current_index; $index <= 9; $index++) {
		if(exists $qd_data->{queue}->{index}->{"$index"}->{project_number} &&
		   $qd_data->{queue}->{index}->{"$index"}->{project_number} =~ /^\d+$/
		) {
			my $prcg  = sprintf("%4d", $qd_data->{queue}->{index}->{"$index"}->{project_number});
			   $prcg .= " (".$qd_data->{queue}->{index}->{"$index"}->{folding}->{run}.",";
			   $prcg .= " ".$qd_data->{queue}->{index}->{"$index"}->{folding}->{clone}.",";
			   $prcg .= " ".$qd_data->{queue}->{index}->{"$index"}->{folding}->{generation}.")";
		
			if($qd_data->{queue}->{index}->{"$index"}->{flops}) {
				if($cfg{verbose}) {
					print "Adding Flops: ";
					print $qd_data->{queue}->{index}->{"$index"}->{flops};
					print "  for: $prcg\n";
				}
			   
				push @project, $prcg;
				push @ppd, $qd_data->{queue}->{index}->{"$index"}->{flops};
			}
		}
	}
	for(my $index = 0; $index < $current_index; $index++) {
		if(exists $qd_data->{queue}->{index}->{"$index"}->{project_number} &&
		   $qd_data->{queue}->{index}->{"$index"}->{project_number} =~ /^\d+$/
		) {
			my $prcg  = sprintf("%4d", $qd_data->{queue}->{index}->{"$index"}->{project_number});
			   $prcg .= " (".$qd_data->{queue}->{index}->{"$index"}->{folding}->{run}.",";
			   $prcg .= " ".$qd_data->{queue}->{index}->{"$index"}->{folding}->{clone}.",";
			   $prcg .= " ".$qd_data->{queue}->{index}->{"$index"}->{folding}->{generation}.")";
		
			if($qd_data->{queue}->{index}->{"$index"}->{flops}) {
				if($cfg{verbose}) {
					print "Adding Flops: ";
					print $qd_data->{queue}->{index}->{"$index"}->{flops};
					print "  for: $prcg\n";
				}
			   
				push @project, $prcg;
				push @ppd, $qd_data->{queue}->{index}->{"$index"}->{flops};
			}
		}
	}

	my $graph_file = $queue_graphs_dir.$client_name.'-flops.png';
	my $width      = 500;   # GoT Max: 570
	my $height     = 400;   # GoT Max: 430

	my @x = @project;
	my @y = @ppd;
	my @l = @project; # legend

	#my $max = 0;
	#foreach(@ppd) {
	#	if($_ > $max) {
	#		$max = $_;
	#	}
	#}
	#$max = int($max + 50);

	if(@x && @y) {
		my $date = get_date();
	
	        my @data  = (\@x, \@y);
	        my $graph = GD::Graph::bars->new($width, $height);
	
	        my $data = GD::Graph::Data->new(\@data);
       		my $values = $data->copy;
		   
		   $graph->set(show_values => $values);
		   $graph->set(dclrs => [ qw(black) ] );
		   
		   $graph->set( x_label           => "Project (Run, Clone, Generation)",
		   		y_label           => "Flops",
				title             => "Flops Of Work Units In The Queue",
				x_labels_vertical => 1,
				#y_max_value       => $max,
				
				fgclr             => '#000000',	
				bgclr             => '#FFFFFF',	
				labelclr          => '#000000',
				axislabelclr      => '#000000',
				legendclr         => '#000000',
				valuesclr         => '#000000',
				textclr           => '#000000',
			      ) or die $graph->error;
			
		my $gd = $graph->plot(\@data) or die $graph->error;
		
		my ($success, $reason) = save_file( $graph_file, 
		                                    $graph->plot(\@data)->png, 
						    mode => "0664", verbose => $cfg{verbose}
						  );
		if($success) {
			# add the graph file for this client to the queue-graphs list
			#$queue_graphs->{queue_graphs}->{client}->{$client_name}->{file} = $graph_file;
			#$queue_graphs->{queue_graphs}->{client}->{$client_name}->{last_update} = $date;

			print "Successfully generated graph for client: $client_name\n" if($cfg{verbose});
		}
		else {
			print $reason;
			exit 1;
		}
	}
	else {
		print "No data found to generate graph for client: $client_name.\n" if($cfg{verbose});
	}
}


sub generate_tpf_graph_for_client {
	my $client_name = shift;
	my %cfg         = @_;
        
	my $qd_file = $client_list->{client_list}->{client}->{$client_name}->{files}->{qd_data};

        # qd-data.xml content
	print "Parsing qd-data XML file: $qd_file ... \n" if($cfg{verbose});
        my $qd_data = read_xml( %cfg, 
				Name => 'qd',
				File => $qd_file
			      ) || die "Failed to read project-images XML file!\n";

	#print Dumper($qd_data);
	
	my $current_index = -1;
	#print "CI:$current_index\n";
	if($qd_data->{queue}->{current_index} ne '') {
		$current_index = $qd_data->{queue}->{current_index};
	}
	#print "CI:$current_index\n";
	
	# Start with the oldest WU
	#if($current_index == 9) {
	#	$current_index = 0;
	#}
	#else {
	#	$current_index += 1;
	#}

	if($client_list->{client_list}->{client}->{$client_name}->{files}->{fahlog} &&
	   -r $client_list->{client_list}->{client}->{$client_name}->{files}->{fahlog}
	) {
		my $project       = -1;
		my $run		  = -1;
		my $clone         = -1;
		my $generation    = -1;

		if(exists $qd_data->{queue}->{index}->{$current_index}->{project_number}            &&
		   exists $qd_data->{queue}->{index}->{$current_index}->{folding}->{run}            &&
		   exists $qd_data->{queue}->{index}->{$current_index}->{folding}->{clone}          &&
		   exists $qd_data->{queue}->{index}->{$current_index}->{folding}->{generation}     &&
		   $qd_data->{queue}->{index}->{$current_index}->{project_number}        =~ /^\d+$/ &&
		   $qd_data->{queue}->{index}->{$current_index}->{folding}->{run}        =~ /^\d+$/ &&
		   $qd_data->{queue}->{index}->{$current_index}->{folding}->{clone}      =~ /^\d+$/ &&
		   $qd_data->{queue}->{index}->{$current_index}->{folding}->{generation} =~ /^\d+$/
		) {
			$project    = $qd_data->{queue}->{index}->{$current_index}->{project_number};
			$run        = $qd_data->{queue}->{index}->{$current_index}->{folding}->{run};
			$clone      = $qd_data->{queue}->{index}->{$current_index}->{folding}->{clone};
			$generation = $qd_data->{queue}->{index}->{$current_index}->{folding}->{generation};

			$run        = 0 if(!$run);
			$clone      = 0 if(!$clone);
			$generation = 0 if(!$generation);

			print "Project: $project\n";
			print "Run    : $run\n";
			print "Clone  : $clone\n";
			print "Gen    : $generation\n";
		}
		else {
			print "Only Folding\@Home unit are supported, but there were no\n";
			print "Project, Run, Clone & Generation numbers found in qd-data.xml.\n";

			return;
		}
		
		my $i = 0;
		my @steps;
		my %seen;
		my $total;
		
		my $time;
		my $steps_completed;
		my $steps_total;
		my $percentage;

		my $found_project = 0;
		my $prcg          = '';
		
		my $file = $client_list->{client_list}->{client}->{$client_name}->{files}->{fahlog};
                open(F, $file) || die "Cannot open log file: $file\n";
                while(<F>) {
                        chomp;

			# project string to search for:
			#
			# [13:30:42] Project: 1496 (Run 24, Clone 10, Gen 4)
			#
			# or for the TINKER core:
			#
			# [20:51:00] Protein: p1152_L939_K12M_ext_from638
			# [20:51:00] - Run: 76 (Clone 46, Gen 2)
			# [20:51:00] - Frames Completed: 46, Remaining: 354
			# [20:51:00] - Dynamic steps required: 4425000
			if(!$found_project) {
				if(/^\[(\d{2}:\d{2}:\d{2})\] Project: $project \(Run $run, Clone $clone, Gen $generation\)$/) {
					$found_project = 1;
					$prcg = "$project ($run, $clone, $generation)";
	                                
					if($cfg{verbose}) {
	                                        print "Debug: Found project $prcg\n";
	                                }
				}
				elsif($prcg eq '' && /^\[(\d{2}:\d{2}:\d{2})\] Protein: p${project}[a-zA-Z\ \_]/) {
					$prcg = $project;
				}
				elsif($prcg eq $project && /^\[(\d{2}:\d{2}:\d{2})\] - Run: $run \(Clone $clone, Gen $generation\)$/) {
					$found_project = 1;
					$prcg .=  " ($run, $clone, $generation)";
	                                
					if($cfg{verbose}) {
	                                        print "Debug: Found project $prcg\n";
	                                }
				}
			}
			else {
				# Gromacs based cores and Amber/PMD core:
                        	#
				# [09:16:00] Completed 311670 out of 500000 steps  (62)
                        	# [10:36:55] Completed 315000 out of 500000 steps  (63)
				#
                        	# [09:16:00] Completed 311670 out of 500000 steps  (62%)
                        	# [10:36:55] Completed 315000 out of 500000 steps  (63%)
                        	#
				# [09:16:00] Completed 311670 out of 500000 steps  (62 percent)
                        	# [10:36:55] Completed 315000 out of 500000 steps  (63 percent)
	                        #
				if(/^\[(\d{2}:\d{2}:\d{2})\] Completed (\d+) out of (\d+) steps  \((\d+).*?\)$/) {
	                                $time            = $1;
	                                $steps_completed = $2;
	                                $steps_total     = $3;
	                                $percentage      = $4;
	
	                                if($cfg{verbose}) {
	                                        print "Debug: Time: $time; Steps completed: $steps_completed, total: $steps_total; Percentage: $percentage\n";
	                                }
	                                # if this is a previously seen step, skip it
	                                if($seen{$steps_completed}) {
	                                        print "Already seen step: $steps_completed (". ($i-1) ."), skipping.\n" if($cfg{verbose});
	                                        next;
	                                }
	
	
	                                push @steps, { time       => $time,
	                                               completed  => $steps_completed,
        	                                       total      => $steps_total,
        	                                       percentage => $percentage
        	                                     };
	
	                                # flag this step as seen
	                                $seen{$steps_completed} = $time;
	
	                                $i++;
				}
				# GPU core:
				#
				# [05:39:50] Completed 10
				#
				# GPU work/logfile_??.txt can contain:
				#
				# Completed 10%
				# 
				elsif(/^\[(\d{2}:\d{2}:\d{2})\] Completed (\d+)(\%|)$/) {
					my $s1 = $1;
					my $s2 = $2;
					my $s3 = $3;
				
	                                $time            = $s1;
	                                $steps_completed = $s2;
	                                $steps_total     = 100;
	                                $percentage      = $s2;
	
	                                if($cfg{verbose}) {
	                                        print "Debug: Time: $time; Steps completed: $steps_completed, total: $steps_total; Percentage: $percentage\n";
	                                }
	                                # if this is a previously seen step, skip it
	                                if($seen{$steps_completed}) {
	                                        print "Already seen step: $steps_completed (". ($i-1) ."), skipping.\n" if($cfg{verbose});
	                                        next;
	                                }
	
	
	                                push @steps, { time       => $time,
	                                               completed  => $steps_completed,
        	                                       total      => $steps_total,
        	                                       percentage => $percentage
        	                                     };
	
	                                # flag this step as seen
	                                $seen{$steps_completed} = $time;
	
	                                $i++;
				}
				# TINKER core:
				#
				# [12:37:33] Protein: p1152_L939_K12M_ext_from638
				# [12:37:33] - Run: 43 (Clone 14, Gen 22)
				# [12:37:33] - Frames Completed: 5, Remaining: 395
				# [12:37:33] - Dynamic steps required: 4937500
				# ...
				# [17:28:23] Finished a frame (12)
				# [18:10:00] Finished a frame (13)
	                        #
				elsif(/Frames Completed: (\d+), Remaining: (\d+)$/ ||
	                              /^\[(\d{2}:\d{2}:\d{2})\] Finished a frame \((\d+)\)$/
				) {
	                                my $s1 = $1;
	                                my $s2 = $2;
	
	                                print "Debug: S1: $s1  S2: $s2\n" if($cfg{verbose});
	
	                                # The first regex match is on a sigle digit,
	                                # so we're parsing a "Frames Completed: ..." line
	                                if($s1 =~ /^\d+$/) {
	                                        $steps_completed = $s1;
	                                        $steps_total     = $s1 + $s2;
	                                }
	                                # the first regex match is a timestamp
	                                # so we're parsing a "[00:00:00] Finished a frame (0)" line
	                                elsif($s1 =~ /^\d{2}:\d{2}:\d{2}$/) {
	                                        $time            = $s1;
	                                        $steps_completed = $s2;
	
	                                        # if this is a previously seen step, skip it
	                                        if($seen{$steps_completed}) {
	                                                print "Already seen step: $steps_completed (". ($i-1) ."), skipping.\n" if($cfg{verbose});
	                                                next;
	                                        }
	
	                                        push @steps, { time       => $time,
							       completed  => $steps_completed,
							       total      => $steps_total,
							       percentage => $percentage
							     };

        	                                # flag this step as seen
        	                                $seen{$steps_completed} = $time;
	
	                                        $i++;
	
	                                        last if($steps_completed == $steps_total);
	                                }

       		                 }
			}
		}
		close F;

                print "Found $i lines with progress information.\n" if($cfg{verbose});
		
		if($i > 0) {
	                # calculate the times between the steps
	                #
	                # How?: Get the last seen time, and the current time,
	                #       compare both, while assuming the current time is for
	                #       today, and the previous time might be yesterday or today
	                #       (maybe the log spans multiple days).
	                #       Calculate the difference between those points in time
	                #       and add that to the list of times.
	
	                $i = 0;
	
	                my $previous = 0;
	                my $current  = 0;
	
	                my $err;
	                my $mode = 1;
	                my $delta;
	
			my @times;
			
			foreach(@steps) {
				# skip the first time as we can't compare it to his predecessor
	                        if($i == 0) {
	                                $i++;
	                                next;
	                        }
	
	                        $previous = $steps[$i-1]->{time};
	                        $current  = $steps[$i]->{time};
	
	                        my $date1 = ParseDate("today ".$previous);
	                        my $date2 = ParseDate("today ".$current);
	                        my $flag  = Date_Cmp($date1, $date2);
	
	                        if($flag < 0) {
	                                # date1 is earlier, the previous is older than the current
	                                #
	                                # the situation may have been:
	                                #  date1:     today 00:01:02 (previous: 00:01:02)
	                                #  date2:     today 00:02:03 (current : 00:02:03)
	                        }
	                        elsif($flag == 0) {
	                                # the two dates are identical
	                        }
	                        else {
	                                # date2 is earlier, the current is older than the previous
	                                # the situation may have been:
	                                #  date1: yesterday 23:59:58 (previous: 23:59:58)
	                                #  date2:     today 00:01:02 (current : 00:01:02)
	
        	                        $date1 = ParseDate("yesterday ".$previous);
	
	                                if($cfg{verbose}) {
	                                        print "Warning: The current date in the list is earlier than the previous ($current < $previous)\n";
	                                        print "The previous date was probably yesterday at $previous.\n";
	                                }
	
	                        }
	                        print "Parsing dates: $date1 and $date2 ($i).\n" if($cfg{verbose});

        	                $delta = DateCalc($date1, $date2, \$err, $mode);
	
        	                print "Difference between $date1 and $date2 is $delta.\n" if($cfg{verbose});
	
	
	                        # add number of seconds in difference to the list of times per step
	                        my $seconds = -1;
	
	                        #+0:0:0:0:1:55:43
	                        if($delta =~ /(\d+):(\d+):(\d+):(\d+):(\d+):(\d+):(\d+)$/) {
	                                #+Y:+M:+W:+D:+H:+MN:+S
	                                my $years   = $1;
	                                my $months  = $2;
	    	                        my $weeks   = $3;
	                                my $days    = $4;
	                                my $hours   = $5;
	                                my $minutes = $6;
	                                   $seconds = $7;
	
	                                if($cfg{verbose}) {
	                                        print "Debug: ";
	                                        print "years: $years ";
	                                        print "months: $months ";
	                                        print "weeks: $weeks ";
	                                        print "days: $days ";
	                                        print "hours: $hours ";
	                                        print "minutes: $minutes ";
	                                        print "seconds: $seconds\n";
	                                }
	
	                                $seconds += ($minutes * 60);                         #       60   seconds per minute
	                                $seconds += ($hours * 60 * 60);                      #     3600   seconds per hour
	                                $seconds += ($days * 24 * 60 * 60);                  #    86400   seconds per day
	                                $seconds += ($months * 30.41667 * 24 * 60 * 60);     #   262800.3 seconds per month
	                                $seconds += ($years * 12 * 30.41667 * 24 * 60 * 60); # 31536003   seconds per year
	                        }
	                        print "Debug: Seconds: $seconds\n" if($cfg{verbose});
	                        print "Debug: Minutes: ".($seconds/60)."\n" if($cfg{verbose});
	                        #push @times, sprintf("%.0f", $seconds/60);
	                        push @times, $seconds;
	
	                        # date1            - date2
	                        # 2006062523:45:28 - 2006062601:41:11 | +0:0:0:0:1:55:43.
	                        #              43 sec +
	                        #  55 min  = 3300 sec +
	                        #   1 hour = 3600 sec =
	                        #            6943 sec
	
	                        $i++;
	                }
			
	
			my @x = (); # frame number
			my @y = (); # time in seconds
			
			#unshift @times, (undef, undef);
	
			@x = 1..$#times;
			@y = @times;
			
			my $min = '';
			my $max = 0;
			my $tot = 0;
			my $avg = 0;
			   $i   = 0;
			print "The Times:\n";
			foreach(@times) {
			        print "$i: $_ ";
				print time_format($_);
				print "\n";
					
				if($_ > $max) {
					$max = $_;
				}
				elsif($min eq '' || ($_ < $min)) {
					$min = $_;
				}
				
				$tot += $_;
				
				$i++;
			}
			$avg = sprintf("%.0f", $tot / $i);
		
			print "Max: $max ";
			print time_format($max);
			print "\n";
			print "Min: $min ";
			print time_format($min);
			print "\n";
			print "Tot: $tot ";
			print time_format($tot);
			print "\n";
			print "Avg: $avg ";
			print time_format($avg);
			print "\n";

			if(@x && @y) {
		        	my $graph_file = $queue_graphs_dir.$client_name.'-tpf.png';
		        	my $width      = 570;   # GoT Max: 570
		        	my $height     = 430;   # GoT Max: 430
	
				my $date = get_date();
		
				my @data  = (\@x, \@y);
				my $graph = GD::Graph::lines->new($width, $height);
			   	   $graph->set( x_label           => "Frame",
				                y_label           => "Time",
				                title             => "Time per Frame for: $prcg",
						
						y_max_value       => $max,
						y_min_value       => $min,
	
						x_min_value       => 0,
						x_max_value       => $#times,
	
						x_label_skip      => 1,
						y_label_skip      => 5,
						y_tick_number     => ($max - $min),
						y_number_format   => \&time_format,
						
						values_format     => \&time_format,
				                #values_vertical   => 1,
						
						skip_undef        => 1,
						
						fgclr             => '#000000',	
						bgclr             => '#FFFFFF',	
						labelclr          => '#000000',
						axislabelclr      => '#000000',
						legendclr         => '#000000',
						valuesclr         => '#000000',
						textclr           => '#000000',
						dclrs             => [ '#FF0000' ],
			              ) || die $graph->error;
				
				my $data = GD::Graph::Data->new(\@data);
				my $values = $data->copy;
				
				   $graph->set(show_values => $values);
					
				my $gd = $graph->plot(\@data) or die $graph->error;
	
				my ($success, $reason) = save_file( $graph_file,
								    $graph->plot(\@data)->png,
								    mode => "0664", verbose => $cfg{verbose}
								  );
				
				if($success) {
					# add the graph file for this client to the queue-graphs list
					#$queue_graphs->{queue_graphs}->{client}->{$client_name}->{file} = $graph_file;
					#$queue_graphs->{queue_graphs}->{client}->{$client_name}->{last_update} = $date;
					
					print "Successfully generated graph for client: $client_name\n" if($cfg{verbose});
				}
				else {
					print $reason;
					exit 1;
				}
			}
		}
	}	
}

sub has_qd_data {
	my $client      = shift;
	my $client_list = shift;
	my %cfg         = @_;

	if(!exists $client_list->{client_list}->{client}->{$client}->{files}->{qd_data}) {
		if($cfg{verbose}) {
			print "No qd-data.xml listed in the client-list file.\n";
		}
		return 0;
	}
	elsif($client_list->{client_list}->{client}->{$client}->{files}->{qd_data} =~ /^\s*$/) {
		if($cfg{verbose}) {
			print "Warning: Empty qd-data listed.\n";
		}
		return 0;
	}
	elsif(!-r $client_list->{client_list}->{client}->{$client}->{files}->{qd_data}) {
		if($cfg{verbose}) {
			print "Warning: Cannot read qd-data.xml file: ";
			print $client_list->{client_list}->{client}->{$client}->{files}->{qd_data};
			print "\n";
		}
		return 0;
	}

	return 1;
}

sub has_fahlog {
	my $client      = shift;
	my $client_list = shift;
	my %cfg         = @_;

	if(!exists $client_list->{client_list}->{client}->{$client}->{files}->{fahlog}) {
		if($cfg{verbose}) {
			print "No FAHlog.txt listed in the client-list file.\n";
		}
		return 0;
	}
	elsif($client_list->{client_list}->{client}->{$client}->{files}->{fahlog} =~ /^\s*$/) {
		if($cfg{verbose}) {
			print "Warning: Empty FAHlog.txt listed.\n";
		}
		return 0;
	}
	elsif(!-r $client_list->{client_list}->{client}->{$client}->{files}->{fahlog}) {
		if($cfg{verbose}) {
			print "Warning: Cannot read FAHlog.txt file: ";
			print $client_list->{client_list}->{client}->{$client}->{files}->{fahlog};
			print "\n";
		}
		return 0;
	}

	return 1;
}

sub time_format {
	my $value = shift;
	my $ret;
	
	if ($value >= 0)
	{
		# convert time in seconds to human readable format
		# e.g. 65 seconds == 1m5s
		
		my $date1 = ParseDate("today");
		my $date2 = DateCalc($date1, "+ $value seconds");
		
		my $delta = DateCalc($date1, $date2);
		#print "Delta: $delta\n";

		#+0:0:0:0:1:55:43
		if($delta =~ /(\d+):(\d+):(\d+):(\d+):(\d+):(\d+):(\d+)$/) {
			#+Y:+M:+W:+D:+H:+MN:+S
			my $years   = $1;
			my $months  = $2;
			my $weeks   = $3;
			my $days    = $4;
			my $hours   = $5;
			my $minutes = $6;
			my $seconds = $7;
			
			my $time = '';
			if($years > 0) {
				$time .= $years."y";
			}
			if($months > 0) {
				$time .= $months."m";
			}
			if($days > 0) {
				$time .= $days."d";
			}
			if($hours > 0) {
				$time .= $hours."h";
			}
			if($minutes > 0) {
				$time .= $minutes."m";
			}
			if($seconds > 0) {
				$time .= $seconds."s";
			}

			if($cfg{verbose}) {
				#print "Debug: ";
				#print "years: $years ";
				#print "months: $months ";
				#print "weeks: $weeks ";
				#print "days: $days ";
				#print "hours: $hours ";
				#print "minutes: $minutes ";
				#print "seconds: $seconds\n";
				#
				#print "time: $time\n";
			}

			$ret = $time;
		}
		else {
			$ret = $value;
		}
	}
	else
	{
		$ret = $value;
	}
	
	
	return $ret;
}

sub print_usage {
	my %cfg = @_;
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --dir <DIR>         Path of the FCI server directory\n";
	print "-c, --client <NAME>     Client to generate graphs for\n";
	
	print "-v, --verbose           Enable verbose output\n";
	print "-q, --quiet             Don't print any output at all, except for errors\n";
	print "-h, --help              Show this usage information\n";
}
